/*
    FEARLUS model0-7-1: Environment.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the environment class.

*/

#import "Environment.h"
#import "BitString.h"
#import "LandParcel.h"
#import "LandUse.h"
#import "MiscFunc.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "EnvironmentHistory.h"
#import "ClassInfo.h"
#import "Coordinate2D.h"
#import "Interval.h"
#import <string.h>
#import <errno.h>
#import <objc/objc-api.h>
#import <objc/Protocol.h>

#define EOFCHECK(expr, filen, file) if(!(expr)) { \
  fprintf(stderr, "%s -- Unexpected EOF in file %s\n", sel_get_name(_cmd), \
          (filen)); \
  [(file) drop]; \
  return self; \
}

/*

getClassMatch(TypeString, OptionalSuffix, Protocol, Separator, WhichOne)

Called by getTopologyClass and getNeighbourhoodClass to return a class
matching the required type. Separator is the separator in the typeString
between the topology and the neighbourhood, and WhichOne is whether to take
the one before or after the Separator (0 for before, 1 for after).

*/

static Class getClassMatch(char *typeStr, const char *suffix, 
			   Protocol *proto, char separator, int whichOne) {
  char *type = strdup(typeStr);
  char *str;
  char *longStr;
  char *p, *q;
  Class theClass = Nil;
  id <List> classList;

  if(type == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  p = strchr(type, (int)'-');
  if(p == NULL) {
    fprintf(stderr, "Invalid format in type string\n");
    return Nil;
  }

  *p = '\0';
  if(whichOne == 0) {
    q = type;		/* Set q to point to the topology class */
  }
  else {
    q = p + 1;		/* Set q to point to the neighbourhood class */
  }
  str = strdup(q);
  if(str == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  longStr = (char *)malloc(strlen(str) + strlen(suffix) + 1);
  if(longStr == NULL) {
    perror("Memory allocation");
    exit(1);
  }

  sprintf(longStr, "%s%s", str, suffix);

  classList = [List create: scratchZone];
  [ClassInfo getClassesForProtocol: proto inList: classList];
  
  while([classList getCount] > 0) {
    Class curClass = (Class)[classList removeFirst];

    if(strcmp([curClass name], str) == 0) {
      theClass = curClass;			// Return if exact match
      break;
    }
    else if(strcmp([curClass name], longStr) == 0) {
      theClass = curClass;		/* Save in case get exact match later
					   (like some fool's going to create
					   a class called XXXTopologyTopology)
					*/
    }
  }
  if(theClass == Nil) {
    fprintf(stderr, "No class found for %s or %s%s\n", str, str, suffix);
  }
  free(type);
  free(str);
  free(longStr);
  return theClass;
}

/*

getTopologyClass(EnvironmentTypeString)

Return the Class of the topology of the environment from the environment type
string. The environment type string has the form <TOPOLOGY>-<NEIGHBOURHOOD>
(e.g. "Toroidal-Moore" or "ToroidalTopology-MooreNeighbourhood"). This function
looks in the list of classes available that follow the Topology protocol for
a class with either the same name as the TOPOLOGY bit of the string, or with
"Topology" appended onto the end of the string.

*/

static Class getTopologyClass(char *typeStr) {
  Class topologyClass = getClassMatch(typeStr, "Topology",
				      @protocol(Grid2DTopology), '-', 0);
  if(topologyClass == Nil) {
    fprintf(stderr, "Invalid environment type string: %s -- it must be of the "
	    "format <TOPOLOGY>-<NEIGHBOURHOOD>, where <TOPOLOGY> is the name "
	    "of a class following the Topology protocol, and <NEIGHBOURHOOD> "
	    "is likewise for the Neighbourhood protocol\n", typeStr);
    exit(1);
  }
  return topologyClass;
}

/*

getNeighbourhoodClass(EnvironmentTypeString)

Return the Class of the neighbourhood from the environment type string, as
for topology...

*/

static Class getNeighbourhoodClass(char *typeStr) {
  Class nbrClass = getClassMatch(typeStr, "Neighbourhood",
				 @protocol(Neighbourhood), '-', 1);

  if(nbrClass == Nil) {
    fprintf(stderr, "Invalid environment type string: %s -- it must be of the "
	    "format <TOPOLOGY>-<NEIGHBOURHOOD>, where <TOPOLOGY> is the name "
	    "of a class following the Topology protocol, and <NEIGHBOURHOOD> "
	    "is likewise for the Neighbourhood protocol\n", typeStr);
    exit(1);
  }
  return nbrClass;
}

  

@implementation Environment

/*

setParametersClimateBitLength:economyBitLength:landParcelBitLength:.....

Use one method to set all the model parameters that apply to the environment.
This saves overheads on calling methods. (But for all I know, probably creates
overheads because of the number of arguments.)

*/

-(void)setParameters: (Parameter *)p {
  Coordinate2D *size;

  parameter = p;
  sClimate = [parameter climateBitStringSize];
  sEconomy = [parameter economyBitStringSize];
  sLandParcel = [parameter parcelBitStringSize];
  sLandUse = [parameter useBitStringSize];
  nLandUse = [parameter nUses];
  size = [Coordinate2D create: scratchZone setX: [parameter environmentXSize]
		       Y: [parameter environmentYSize]];
  [self setSizeX: (int)size->x Y: (int)size->y];
				// N.B. This may not work, because the setSizeX
				// method is for the Creating phase only.
				// Well, it will work if this method is called
				// between createBegin and createEnd
  pLUDontCare = [[Interval create] set: [parameter pUseDontCareBit]];
  year = 0;
  landParcels = [List create: [self getZone]];
  landUses = [Array create: [self getZone] setCount: (int)nLandUse];
  climate = [BitString create: [self getZone] setBitCount: sClimate];
  [climate setEachBitTo: 1 withProbability: [[Interval constant: 0.5] call]];
  economy = [BitString create: [self getZone] setBitCount: sEconomy];
  [economy setEachBitTo: 1 withProbability: [[Interval constant: 0.5] call]];
  history = [List create: [self getZone]];
  n_centre = [Coordinate2D create: [self getZone] setX: -1 Y: -1];
  topology
    = [getTopologyClass([parameter environmentType])
		       create: [self getZone]
		       setSize: size
		       nbrClass:
			 getNeighbourhoodClass([parameter environmentType])
		       radius: [parameter neighbourhoodRadius]];
  [size drop];
}

/*

setMaxColour

Set the maximum colour value that has been successfully allocated

*/

-(void)setMaxColour: (int)col {
  maxColour = col;
}

/*

calculateYield

Calculate the yield of each land parcel

*/

-(void)calculateYield {
  int x, y;

  if([FearlusArguments debugEnvironment]) {
    printf("E: Calculating yield for each land parcel\n");
  }
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      [[self getObjectAtX: x Y: y] calculateYield];
    }
  }
}

/*

determineClimate

A method for computing the climate. This may either involve loading
the climate from a file, or, if not, toggling each bit of the climate
with specified toggle probabilities -- one for each bit. If the
probability is 0.5 for each bit, then effectively the climate is
random, as per model0-2. Note that when the climate object is created,
it is set to random values at that time. If the climate file is
specfied but does not exist, then the climate is computed using the
toggle array, as it would be if no file were specified, however, it is
then saved to the climate file.

*/

-(void)determineClimate {
  if([parameter useClimateFile]) {
    if(climateFileStatus == NOT_OPENED_YET) {
      // Initialise the climate file pointer
      if([MiscFunc fileExists: [parameter climateFile]]) {
	climateFile = fopen([parameter climateFile], "r");
	if(climateFile == NULL) {
	  fprintf(stderr, "%s -- could not open climate file %s for reading",
		  sel_get_name(_cmd), [parameter climateFile]);
	  perror("");
	  exit(1);
	}
	climateFileStatus = OPENED_FOR_READING;
      }
      else {
	climateFile = fopen([parameter climateFile], "w");
	if(climateFile == NULL) {
	  fprintf(stderr, "%s -- could not create climate file %s",
		  sel_get_name(_cmd), [parameter climateFile]);
	  perror("");
	  exit(1);
	}
	climateFileStatus = OPENED_FOR_WRITING;
      }
    }
    // Load in, or save the data. Deliberately not using an else if, because we
    // want to continue on from the initial status NOT_OPENED_YET, if that was
    // the case.
    if(climateFileStatus == OPENED_FOR_READING) {
      // We're in read mode
      char *buf;
      int i;

      buf = (char *)malloc([parameter climateBitStringSize] + 2);
      if(buf == NULL) {
	perror("Memory allocation");
	exit(1);
      }
      for(i = 0; i < [parameter climateBitStringSize]; i++) {
	buf[i] = fgetc(climateFile);
	if(!(buf[i] == '1' || buf[i] == '0')) {
	  if(feof(climateFile)) {
	    fclose(climateFile);
	    climateFileStatus = CLOSED;
	    if([FearlusArguments debugEnvironment]) {
	      printf("E: No more data from climate file %s. Will continue by "
		     "toggling bits in the climate bit string as per the "
		     "toggle array\n", [parameter climateFile]);
	    }
	    break;
	  }
	  else {
	    fprintf(stderr, "%s -- unexpected character in climate file %s. "
		    "Expecting 1 or 0, found %c\n", sel_get_name(_cmd),
		    [parameter climateFile], buf[i]);
	    exit(1);
	  }
	}
      }
      if(i == [parameter climateBitStringSize]) {
	buf[i] = '\0';
	[climate setToString: buf];
	(void)fgetc(climateFile); 	// Skip over the carriage return
      }
      if([climate getBitCount] != [parameter climateBitStringSize]) {
	fprintf(stderr, "%s -- Invalid format in climate file %s, bit string"
		" loaded has %u bits -- expecting %u\n",
		sel_get_name(_cmd), [parameter climateFile],
		[climate getBitCount], [parameter climateBitStringSize]);
	exit(1);
      }
      free(buf);
    }
    else if(climateFileStatus == OPENED_FOR_WRITING) {
      [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
      [climate printToFile: climateFile];
      fprintf(climateFile, "\n");
      fflush(climateFile);
    }
    // Deal with the file having run out of data. Deliberately not using an
    // else if here, because the status might have been set to CLOSED during
    // reading, and no climate have been read.
    if(climateFileStatus == CLOSED) {
      [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
    }
  }
  else {
    [climate toggleEachBitUsingProbs: [parameter climateToggleProbs]];
  }
  if([FearlusArguments debugEnvironment]) {
    printf("E: New climate: ");
    [climate print];
  }
}

/*

getClimate

An accessor method for the climate.

*/

-getClimate {
  return climate;
}

/*

determineEconomy

A method for computing the economy. Currently this is a random bit string.

*/

-(void)determineEconomy {
  if([parameter useEconomyFile]) {
    if(economyFileStatus == NOT_OPENED_YET) {
      // Initialise the economy file pointer
      if([MiscFunc fileExists: [parameter economyFile]]) {
	economyFile = fopen([parameter economyFile], "r");
	if(economyFile == NULL) {
	  fprintf(stderr, "%s -- could not open economy file %s for reading",
		  sel_get_name(_cmd), [parameter economyFile]);
	  perror("");
	  exit(1);
	}
	economyFileStatus = OPENED_FOR_READING;
      }
      else {
	economyFile = fopen([parameter economyFile], "w");
	if(economyFile == NULL) {
	  fprintf(stderr, "%s -- could not create economy file %s",
		  sel_get_name(_cmd), [parameter economyFile]);
	  perror("");
	  exit(1);
	}
	economyFileStatus = OPENED_FOR_WRITING;
      }
    }
    // Load in, or save the data. Deliberately not using an else if, because we
    // want to continue on from the initial status NOT_OPENED_YET, if that was
    // the case.
    if(economyFileStatus == OPENED_FOR_READING) {
      // We're in read mode
      char *buf;
      int i;

      buf = (char *)malloc([parameter economyBitStringSize] + 2);
      if(buf == NULL) {
	perror("Memory allocation");
	exit(1);
      }
      for(i = 0; i < [parameter economyBitStringSize]; i++) {
	buf[i] = fgetc(economyFile);
	if(!(buf[i] == '1' || buf[i] == '0')) {
	  if(feof(economyFile)) {
	    fclose(economyFile);
	    economyFileStatus = CLOSED;
	    if([FearlusArguments debugEnvironment]) {
	      printf("E: No more data from economy file %s. Will continue by "
		     "toggling bits in the economy bit string as per the "
		     "toggle array\n", [parameter economyFile]);
	    }
	    break;
	  }
	  else {
	    fprintf(stderr, "%s -- unexpected character in economy file %s. "
		    "Expecting 1 or 0, found %c\n", sel_get_name(_cmd),
		    [parameter economyFile], buf[i]);
	    exit(1);
	  }
	}
      }
      if(i == [parameter economyBitStringSize]) {
	buf[i] = '\0';
	[economy setToString: buf];
	(void)fgetc(economyFile); 	// Skip over the carriage return
      }
      if([economy getBitCount] != [parameter economyBitStringSize]) {
	fprintf(stderr, "%s -- Invalid format in economy file %s, bit string"
		" loaded has %u bits -- expecting %u\n",
		sel_get_name(_cmd), [parameter economyFile],
		[economy getBitCount], [parameter economyBitStringSize]);
	exit(1);
      }
      free(buf);
    }
    else if(economyFileStatus == OPENED_FOR_WRITING) {
      [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
      [economy printToFile: economyFile];
      fprintf(economyFile, "\n");
      fflush(economyFile);
    }
    // Deal with the file having run out of data. Deliberately not using an
    // else if here, because the status might have been set to CLOSED during
    // reading, and no economy have been read.
    if(economyFileStatus == CLOSED) {
      [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
    }
  }
  else {
    [economy toggleEachBitUsingProbs: [parameter economyToggleProbs]];
  }
  if([FearlusArguments debugEnvironment]) {
    printf("E: New economy: ");
    [economy print];
  }
}

/*

getEconomy

An accessor method for the economy.

*/

-getEconomy {
  return economy;
}

/*

getLandParcels

An accessor method for the list of land parcels

*/

-getLandParcels {
  return landParcels;
}

/*

createLandUses

A method for creating an array of land uses. These are all random bit strings.
Also initialise the colours. 



*/

-(void)createLandUses {
  int i, /* colourTaken[216], */ minCol, nCol /*, pass, j */;
  id lu;

  if(maxColour < 40) {		// This is terrible!! 40 should be #defined
				// somewhere.
    minCol = 1;
    nCol = maxColour;
  }
  else {
    minCol = 40;
    nCol = maxColour - 39;
  }
  if([FearlusArguments debugEnvironment]) {
    printf("E: %d colours for %u land uses\n", nCol, nLandUse);
  }
//    for(pass = 1; pass < nCol / 2; pass *= 2) {}
//    j = 0;
//    while(pass > 0) {
//      for(i = pass; i < nCol; i += pass) {
//        if(i % (pass * 2) != 0) {
//  	colourTaken[j++] = i + minCol;
//        }
//      }
//      pass /= 2;
//                             	// Create an array of the colours to use
//    }
  
  for(i = 0; i < (int)nLandUse; i++) {
    lu = [LandUse create: [self getZone]];
    [lu setParameters: parameter];	// setParameters: must be called before
                                        // initialiseWithColour:environment:
    [lu initialiseWithColour: minCol + (i % nCol) environment: self];
    [lu randomiseBitsWithDontCare: [pLUDontCare call]];
    [lu updateParcelPortions];
    [landUses atOffset: i put: lu];
    if([FearlusArguments debugEnvironment]) {
      printf("E: Created land use %u with match string: ", [lu getPIN]);
      [[lu getMatchBitString] printToFile: stdout];
      printf(" and don't care string: ");
      [[lu getDontCareBitString] print];
    }
  }
  printf("Created Land Uses\n");
}

/*

createLandParcels

Populate the environment with land parcels. This is done in a spiral starting
at the middle. The reason for this is so that if we use the same seed, but
a different environment size, we will get the same land parcels in the middle
(provided we do not clump). N.B. This facility has not been used or tested.

*/

-(void)createLandParcels {
  int step;
  int dir;	// 0 for X, 1 for Y
  int sign;
  int centreX;
  int centreY;
  int x, y;
  int max;
  int nParcels;

  nParcels = [self getSizeX] * [self getSizeY];
  centreX = [self getSizeX] / 2;
  centreY = [self getSizeY] / 2;
  x = 0;
  y = 0;
  sign = 1;
  step = 1;
  dir = 0;
  max = 1;
  while(nParcels > 0) {
    if((centreX + x) >= 0 && (centreX + x) < [self getSizeX]
       && (centreY + y) >= 0 && (centreY + y) < [self getSizeY]) {
      // The spiral will build up a square. We might have a rectangular
      // environment. So only create a land parcel if it is inside the
      // environment.
      [self createLandParcelAtX: centreX + x Y: centreY + y];
      nParcels--;
    }
    if(dir == 0) {
      x += sign;
      if(x == max) {
	dir = 1;
      }
    }
    else {
      y += sign;
      if(y == max) {
	dir = 0;
	sign *= -1;
	step++;
	max = ((step / 2) * sign) + (step % 2);
      }
    }
  }
  printf("Created Land Parcels\n");
}

/*

createLandParcelAtX:Y:

Create a land parcel at the specified co-ordinate

*/

-(void)createLandParcelAtX: (int)x Y: (int)y {
  id lp;

  lp = [LandParcel create: [self getZone]];
  [lp setParameters: parameter];
  [lp setBitLength: sLandParcel];
  [lp randomiseBits];
  [lp setXPos: x YPos: y];
  [lp setEnvironment: self];
  if(maxColour >= 40) {
    [lp setMaxGraphColour: 40];
  }
  else {
    [lp setMaxGraphColour: maxColour];
  }
  [self putObject: lp atX: x Y: y];
  [landParcels addLast: lp];
  if([FearlusArguments debugEnvironment]) {
    printf("E: Created land parcel %u at (%d, %d) with bit string: ",
	   [lp getPIN], x, y);
    [[lp getAttributes] print];
  }
}

/*

getLandParcelAtX:Y:

Get the land parcel at the specified co-ordinate. Implement a wrap-around
effect (so bounded environments beware...)

*/

-(LandParcel *)getLandParcelAtX: (int)x Y: (int)y {
  if(x < 0 || x >= [self getSizeX]) {
    x = [MiscFunc Mod: x inRange0To: [self getSizeX]];
  }
  if(y < 0 || y >= [self getSizeY]) {
    y = [MiscFunc Mod: y inRange0To: [self getSizeY]];
  }
  return (LandParcel *)[self getObjectAtX: x Y: y];
}

/*

clumpLandParcels

Use the parameter clumper object to clump the land parcels

*/

-(void)clumpLandParcels {
  int x, y;

  if([parameter clumper] != nil) {
    [[parameter clumper] clumpEnvironment: self];
    if([FearlusArguments debugEnvironment]) {
      printf("E: New bit strings for land parcels:\n");
      for(x = 0; x < [self getSizeX]; x++) {
	for(y = 0; y < [self getSizeY]; y++) {
	  LandParcel *lp;

	  lp = [self getObjectAtX: x Y: y];
	  printf("\t%u at (%d, %d): ", [lp getPIN], x, y);
	  [[lp getAttributes] print];
	}
      }
    }
  }
}

/*

printTextParcelBitstrings

A method to print the parcel bit strings as text to standard output. Loops 
through each bit and writes an ASCII representation of the grid (one char per
cell) showing the setting of that bit.

*/

-(void)printTextParcelBitstrings {
  unsigned a;

  for(a = 0; a < [parameter parcelBitStringSize]; a++) {
    [self printTextParcelBitstringsBit: a];
  }
}

/*

printTextParcelBitstringsBit:

Print a grid showing the bitstring values for a single bit of the Land Parcel
bit strings

*/

-(void)printTextParcelBitstringsBit: (unsigned)a {
  int x, y;

  printf("E: %s: Bit %u:\n", sel_get_name(_cmd), a);
  for(y = [self getSizeY] - 1; y >= 0; y--) {
    printf("\t   %d ", y % 10);
    for(x = 0; x < [self getSizeX]; x++) {
      if([[self getObjectAtX: x Y: y] getOneAttribute: a] != 0) {
	printf("X");
      }
      else {
	printf(".");
      }
    }
    printf("\n");
    if(y == 0) {
      printf("\ty\n\t^    ");
      for(x = 0; x < [self getSizeX]; x++) {
	printf("%d", x % 10);
      }
      printf("\n");
    }	
  }
  printf("\t|\n\t+->x\t('X'->1, '.'->0)\n\n");
}

/*

getNLandUses

Return the number of Land Uses

*/

-(unsigned)getNLandUse {
  return nLandUse;
}

/*

getLandUses

Return the array of valid land uses

*/

-getLandUses {
  return(landUses);
}

/*

connectLandParcels

This method is called from the initialisation schedule to connect Land Parcels
together. It sets up the neighbourhood list of each Land Parcel.

*/

-connectLandParcels {
  int x, y;
  int xSize, ySize;
  Coordinate2D *lpCoord;

  xSize = [self getSizeX];
  ySize = [self getSizeY];

  lpCoord = [Coordinate2D create: scratchZone];

  for(x = 0; x < xSize; x++) {
    for(y = 0; y < ySize; y++) {
      id <Grid2DSpatial> nbr;
      LandParcel *lp = nil;
      
      lp = (LandParcel *)[self getObjectAtX: x Y: y];
      lpCoord->x = x;
      lpCoord->y = y;
      for(nbr = [topology startSearchWithCentre: lpCoord];
	  [topology continueSearch];	  
	  nbr = [topology nextSearchItem]) {
	if([FearlusArguments debugNeighbourhood]) {
	  printf("Adding (%d, %d) to list of neighbours of (%d, %d)\n",
		 [MiscFunc Mod: [nbr getX] inRange0To: xSize],
		 [MiscFunc Mod: [nbr getY] inRange0To: ySize],
		 [lp getX], [lp getY]);
	}
	[lp addNbr:
	      (LandParcel *)[self getObjectAtX:
				    [MiscFunc Mod: [nbr getX]
					      inRange0To: xSize]
				  Y: [MiscFunc Mod: [nbr getY]
					       inRange0To: ySize]]];
      }
    }
  }
  [lpCoord drop];
  return self;
}


/*

getEconomyBitLength

Return the size of the economy bitstring

*/

-(unsigned)getEconomyBitLength {
  return(sEconomy);
}

/*

getClimateBitLength

Return the size of the climate bitstring

*/

-(unsigned)getClimateBitLength {
  return(sClimate);
}

/*

getLandParcelBitLength

Return the size of the land parcel bitstring

*/

-(unsigned)getLandParcelBitLength {
  return(sLandParcel);
}

/*

getLandUseBitLength

Return the size of the land use bitstring

*/

-(unsigned)getLandUseBitLength {
  return(sLandUse);
}

/*

getTopology

Return the topology object

*/

-(id <Grid2DTopology>)getTopology {
  return topology;
}

/*

newYear

actions to perform in the environment at the beginning of a new year / end
of an old year:

- Make sure the histories of the land parcels and land uses are updated
- Increment the year

*/

-(void)newYear {
  year++;
  printf("--------------------------------------------------------------------"
	 "----------\n");
  printf("New year: %u\n", year);
}

/*

getYear

Return the current year

*/

-(unsigned)getYear {
  return(year);
}

/*

rankLandParcels

Rank the land parcels. Send a message to each land parcel to tell it its rank
in the population of (a) number of changes of land use and (b) number of
changes of land manager. The algorithm is simple. Loop through the land parcels
to find the maximum and minimum value of the variable of interest. Then loop
through them again to set their rank, which is given by:

                   my_value_for_the_variable - minimum
	    1.0 -  -----------------------------------
		            maximum - minimum

It is important that the result is given as a number between 0 and 1, so that
land parcels do not need to know about information like the total number of
land parcels. They need to be able to turn this number straight into a grey
level.

*/

-(void)rankLandParcels {
  id iterator, lp;
  Interval *minLU = [Interval create];
  Interval *minLMgr = [Interval create];
  Interval *maxLU = [Interval create];
  Interval *maxLMgr = [Interval create];
  Interval *myLU = [Interval create];
  Interval *myLMgr = [Interval create];
  Interval *diffLU = [Interval create];
  Interval *diffLMgr = [Interval create];

  iterator = [landParcels begin: scratchZone];
  [iterator next];
  if([iterator getLoc] != Member) {
    [iterator drop];
    return;
  }
  lp = [iterator get];
  [minLU set: [maxLU set: [Interval castInt: [lp getNLUChange]]]];
  [minLMgr set: [maxLMgr set: [Interval castInt: [lp getNLMgrChange]]]];
  for(; [iterator getLoc] == Member;
      [iterator next]) {
    lp = [iterator get];
    [myLU set: [Interval castInt: [lp getNLUChange]]];
    [myLMgr set: [Interval castInt: [lp getNLMgrChange]]];
    [minLU set: ([myLU lt: minLU] ? myLU : minLU)];
    [maxLU set: ([myLU gt: maxLU] ? myLU : maxLU)];
    [minLMgr set: ([myLMgr lt: minLMgr] ? myLMgr : minLMgr)];
    [maxLMgr set: ([myLMgr gt: maxLMgr] ? myLMgr : maxLMgr)];
  }
  [diffLU set: [maxLU diff: minLU]];
  [diffLMgr set: [maxLMgr diff: minLMgr]];
  // At this point, the iterator index should be at the end of the list
  // -- at the special non-member item at the end of the list, in fact,
  // so now iterate back through the list, beginning with a move to the
  // last member of the list
  for([iterator prev]; [iterator getLoc] == Member; [iterator prev]) {
    lp = [iterator get];
    [myLU set: [Interval castInt: [lp getNLUChange]]];
    [myLMgr set: [Interval castInt: [lp getNLMgrChange]]];
    [lp setChangeLURank: 
	  [[[Interval constant: 1.0]
	     diff: [[myLU diff: minLU] quot: diffLU]] call]];
    [lp setChangeLMgrRank:
	  [[[Interval constant: 1.0]
	     diff: [[myLMgr diff: minLMgr] quot: diffLMgr]] call]];
  }
  [iterator drop];

  [minLU drop];
  [minLMgr drop];
  [maxLU drop];
  [maxLMgr drop];
  [myLU drop];
  [myLMgr drop];
  [diffLU drop];
  [diffLMgr drop];
}

/*

recordHistory

Record the history

*/

-(void)recordHistory {
  id iterator;
  LandParcel *lp;
  LandUse *lu;
  EnvironmentHistory *newHd;

  newHd = [EnvironmentHistory create: [parameter historyZone]];
  [newHd initialiseWithYear: year climate: climate economy: economy];
  [history addFirst: newHd];
  if([parameter dumpHistory]
     && [parameter maxHistoryLength] < [history getCount]) {
    [[history removeLast] drop];
  }
  for(iterator = [landParcels begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lp = [iterator get];
    [lp recordHistory];
  }
  [iterator drop];
  for(iterator = [landUses begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lu = [iterator get];
    [lu recordHistory];
  }
  [iterator drop];
}


/*

writeHistory

Tell all the land parcels and land uses to write their histories to a file.

*/

-(void)writeHistory: (FILE *)fp {
  fprintf(fp, "Year\tClimate\tEconomy%s", [FearlusArguments nl]);
  [history forEach: M(printToFile:) : (id)fp];
  fprintf(fp, "Type\tID\tYear\tYield\tLand Use\tLand Manager\tStrategyMethod\t"
	  "X\tY\tAttributes\t%s", [FearlusArguments nl]);
  [landParcels forEach: M(printHistoryToFile:) : (id)fp];
  fprintf(fp, "Type\tID\tYear\tNo of Parcels Assigned\tMatch String\t"
	  "Don't Care String%s", [FearlusArguments nl]);
  [landUses forEach: M(writeLandUseHistoryToFile:) : (id)fp];
}

/*

saveLandParcelsToFileNamed:

Save the land parcels to the file. Actually, a file called name.lp will be
created if the name doesn't have a .lp suffix.

File format:

EnvironmentXSize: xxx
EnvironmentYSize: yyy
LandParcelBitStringLength: nnn
XCoord YCoord BitString
xx0	yy0	bb0
xx1	yy1	bb1
...

*/

-saveLandParcelsToFileNamed: (char *)fname {
  id file;
  char *realfname;
  char *buf;
  int x, y;

  realfname = [MiscFunc getUsableFileNameFrom: fname withSuffix: ".lp"];
  if(realfname == NULL) {
    fprintf(stderr, "%s -- could not create datafile\n", sel_get_name(_cmd));
    return self;
  }
  file = [OutFile create: [self getZone] setName: realfname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not create file ", sel_get_name(_cmd));
    perror(realfname);
    free(realfname);
    return self;
  }
  buf = malloc([parameter parcelBitStringSize] + 2);
  if(buf == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  // 2 is for '\0' at the end and one for luck
  [file putString: "EnvironmentXSize:\t"];
  [file putInt: [self getSizeX]];
  [file putNewLine];
  [file putString: "EnvironmentYSize:\t"];
  [file putInt: [self getSizeY]];
  [file putNewLine];
  [file putString: "LandParcelBitStringLength:\t"];
  [file putUnsigned: [parameter parcelBitStringSize]];
  [file putNewLine];
  [file putString: "XCoord\tYCoord\tBitString\n"];
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      [file putInt: x];
      [file putTab];
      [file putInt: y];
      [file putTab];
      if([parameter parcelBitStringSize] > 0) {
	[[[self getObjectAtX: x Y: y] getAttributes] printToString: buf];
	[file putString: buf];
      }
      else {
	[file putString: "E"];
      }
      [file putNewLine];
    }
  }
  [file drop];
  free(realfname);
  free(buf);
  return self;
}

/*

loadLandParcelsFromFileNamed:

Load the land parcels from the file. See saveLandParcelsToFileNamed:
for a description of the format.



*/

-loadLandParcelsFromFileNamed: (char *)fname {
  id file;
  char *buf;
  char word[100];
  int x, y;
  int dummyXSize, dummyYSize, dummyX, dummyY;
  unsigned dummyBSLength;
  LandParcel *lp;

  if([FearlusArguments debugEnvironment]) {
    printf("E: Loading land parcels from file: %s\n", fname);
  }
  file = [InFile create: [self getZone] setName: fname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not open file ", sel_get_name(_cmd));
    perror(fname);
    return self;
  }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "EnvironmentXSize:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "EnvironmentXSize:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getInt: &dummyXSize], fname, file)
    if(dummyXSize != [self getSizeX]) {
      fprintf(stderr, "%s -- wrong value for EnvironmentXSize in file %s. "
	      "Expecting %d found %d\n", sel_get_name(_cmd), fname,
	      [self getSizeX], dummyXSize);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "EnvironmentYSize:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "EnvironmentYSize:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getInt: &dummyYSize], fname, file)
    if(dummyYSize != [self getSizeY]) {
      fprintf(stderr, "%s -- wrong value for EnvironmentYSize in file %s. "
	      "Expecting %d found %d\n", sel_get_name(_cmd), fname,
	      [self getSizeY], dummyYSize);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandParcelBitStringLength:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandParcelBitStringLength:",
	      word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyBSLength], fname, file)
    if(dummyBSLength != [parameter parcelBitStringSize]) {
      fprintf(stderr, "%s -- wrong value for LandParcelBitStringLength in file"
	      " %s. Expecting %u found %u\n", sel_get_name(_cmd), fname,
	      [parameter parcelBitStringSize], dummyBSLength);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "XCoord") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "XCoord", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "YCoord") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "YCoord", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "BitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "BitString", word);
      [file drop];
      return self;
    }
  buf = malloc([parameter parcelBitStringSize] + 2);
  if(buf == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  // 2 is for '\0' at the end and one for luck
  if([FearlusArguments debugEnvironment]) {
    printf("E: New bit strings for land parcels:\n");
  }
  for(x = 0; x < [self getSizeX]; x++) {
    for(y = 0; y < [self getSizeY]; y++) {
      EOFCHECK([file getInt: &dummyX], fname, file)
	EOFCHECK([file getInt: &dummyY], fname, file)
	EOFCHECK([file getWord: buf], fname, file)
	lp = [self getObjectAtX: dummyX Y: dummyY];
      [[lp getAttributes] setToString: buf];
      if([FearlusArguments debugEnvironment]) {
	printf("\t%u at %d, %d): ", [lp getPIN], dummyX, dummyY);
	[[lp getAttributes] print];
      }
    }
  }
  [file drop];
  free(buf);
  return self;
}

/*

saveLandUsesToFileNamed:

Save the land uses to the file. A file called name.lu will be created if the
name doesn't habe a .lu suffix.

File format:

NumberOfLandUses: xxx
LandUseBitStringLength: nnn
LandUseIndex MatchBitString DontCareBitString
xx0		bbm0		bbdc0
xx1		bbm1		bbdc1
...

The index is the index in the array of land uses (not the land use PIN).

*/

-saveLandUsesToFileNamed: (char *)fname {
  id file;
  char *realfname;
  char *buf;
  int i;

  realfname = [MiscFunc getUsableFileNameFrom: fname withSuffix: ".lu"];
  if(realfname == NULL) {
    fprintf(stderr, "%s -- could not create datafile\n", sel_get_name(_cmd));
    return self;
  }
  file = [OutFile create: [self getZone] setName: realfname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not create file ", sel_get_name(_cmd));
    free(realfname);
    perror(realfname);
    return self;
  }
  buf = malloc([parameter useBitStringSize] + 2);
  if(buf == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  // 2 is for '\0' at the end and one for luck -- which we'll need if the 
  // bitstring size is 0
  [file putString: "NumberOfLandUses:\t"];
  [file putUnsigned: [parameter nUses]];
  [file putNewLine];
  [file putString: "LandUseBitStringLength:\t"];
  [file putUnsigned: [parameter useBitStringSize]];
  [file putNewLine];
  [file putString: "LandUseIndex\tMatchBitString\tDontCareBitString\n"];
  for(i = 0; i < [parameter nUses]; i++) {
    [file putInt: i];
    [file putTab];
    [[[landUses atOffset: i] getMatchBitString] printToString: buf];
    [file putString: buf];
    [file putTab];
    [[[landUses atOffset: i] getDontCareBitString] printToString: buf];
    [file putString: buf];
    [file putNewLine];
  }
  [file drop];
  free(realfname);
  free(buf);
  return self;
}

/*

loadLandUsesFromFileNamed:

Load the land uses from the file. See saveLandUsesToFileNamed: for a
description of the format. As for loadLandParcelsFromFileNamed:, this
assumes that the land uses have already been created.

*/

-loadLandUsesFromFileNamed: (char *)fname {
  id file;
  char *buf;
  char word[100];
  int i;
  unsigned dummyNUses, dummyBSLength, dummyI;
  LandUse *lu;

  if([FearlusArguments debugEnvironment]) {
    printf("E: Loading land uses from file: %s\n", fname);
  }
  file = [InFile create: [self getZone] setName: fname];
  if(file == nil) {
    fprintf(stderr, "%s -- could not open file ", sel_get_name(_cmd));
    perror(fname);
    return self;
  }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "NumberOfLandUses:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "NumberOfLandUses:", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyNUses], fname, file)
    if(dummyNUses != [parameter nUses]) {
      fprintf(stderr, "%s -- wrong value for NumberOfLandUses in file %s. "
	      "Expecting %u, found %u\n", sel_get_name(_cmd), fname,
	      [parameter nUses], dummyNUses);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandUseBitStringLength:") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandUseBitStringLength:",
	      word);
      [file drop];
      return self;
    }
  EOFCHECK([file getUnsigned: &dummyBSLength], fname, file)
    if(dummyBSLength != [parameter useBitStringSize]) {
      fprintf(stderr, "%s -- wrong value for LandUseBitStringLength in file "
	      "%s. Expecting %u, found %u\n", sel_get_name(_cmd), fname,
	      [parameter useBitStringSize], dummyBSLength);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "LandUseIndex") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "LandUseIndex", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "MatchBitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "MatchBitString", word);
      [file drop];
      return self;
    }
  EOFCHECK([file getWord: word], fname, file)
    if(strcmp(word, "DontCareBitString") != 0) {
      fprintf(stderr, "%s -- invalid format in file %s. Expecting %s, found "
	      "%s\n", sel_get_name(_cmd), fname, "DontCareBitString", word);
      [file drop];
      return self;
    }
  buf = malloc([parameter useBitStringSize] + 2);
  if(buf == NULL) {
    perror("Memory allocation");
    exit(1);
  }
  // 2 is for '\0' at the end and one for luck
  if([FearlusArguments debugEnvironment]) {
    printf("E: New bit strings for land uses:\n");
  }
  for(i = 0; i < (int)[parameter nUses]; i++) {
    EOFCHECK([file getInt: &dummyI], fname, file)
      lu = [landUses atOffset: i];
    EOFCHECK([file getWord: buf], fname, file)
      [[lu getMatchBitString] setToString: buf];
    EOFCHECK([file getWord: buf], fname, file)
      [[lu getDontCareBitString] setToString: buf];
    [lu updateParcelPortions];
    if([FearlusArguments debugEnvironment]) {
      printf("\t%u match: ", [lu getPIN]);
      [[lu getMatchBitString] printToFile: stdout];
      printf(" don't care: ");
      [[lu getDontCareBitString] print];
    }
  }
  [file drop];
  free(buf);
  return self;
}

/*

drop

Override the superclass method so that all items created by the environment
are also dropped.

*/

-(void)drop {
  if([FearlusArguments debugNeighbourhood]) {
    int i;

    for(i = 0; i < nbrscreenmax; i++) {
      free(nbrscreen[i]);
    }
    free(nbrscreen);
  }
  [n_centre drop];
  [topology drop];
  [economy drop];
  [climate drop];
  [pLUDontCare drop];
  [landUses deleteAll];
  [landParcels deleteAll];
  [super drop];
}

@end





