/*
    FEARLUS model0-7-1: SelectUseBucket.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the SelectUseBucket object. For now, the scores are held in
the land use objects themselves.



*/

#import "SelectUseBucket.h"
#import "FearlusArguments.h"
#import "LandUse.h"
#import "Parameter.h"
#import "Interval.h"

@implementation SelectUseBucket

/*

create:withParameters:andLandUses:

Return a freshly created SelectUseBucket object, with all the scores of the
land uses initialised to zero.

*/

+create: aZone withParameters: (Parameter *)p andLandUses: lu {
  SelectUseBucket *obj;
  int i;

  obj = [super create: aZone];
  obj->parameter = p;
  obj->landUses = lu;

  for(i = 0; i < (int)[p nUses]; i++) {
    [[lu atOffset: i] initialiseScore];
  }

  return obj;
}

/*

addScore:toBucketForLandUse:

Add a score to the bucket for the specified land use.

*/

-(void)addScore: (Interval *)s toBucketForLandUse: (LandUse *)lu {
  [lu addToScore: [s call]];
  [s drop];
}

/*

addAverageScore:toBucketForLandUse:

Add a score to the bucket for the specified land use, but when we get the
score, we want to know the average score over all the times that land use was
sampled.

*/

-(void)addAverageScore: (Interval *)s toBucketForLandUse: (LandUse *)lu {
  [lu addToScoreAverage: [s call]];
  [s drop];
}

/*

getScoreInBucketOfLandUse:

Return the score in the bucket of the specified land use.

*/

-(Interval *)getScoreInBucketOfLandUse: (LandUse *)lu {
  return [lu getScore];
}

/*

setScore:inBucketForLandUse:

Set a score to the bucket for the specified land use.

*/

-(void)setScore: (Interval *)s inBucketForLandUse: (LandUse *)lu {
  [lu setScore: [s call]];
  [s drop];
}

/*

getChoice

Make a choice at random, weighted by the values in the bucket. Return the
result.

*/

-(LandUse *)getChoice {
  int i;
  double choice;
  LandUse *lu;

  /* First we have to deal with average scores. They must be converted to
     normal scores */
  for(i = 0; i < [parameter nUses]; i++) {
    Interval *score = [Interval create];

    lu = (LandUse *)[landUses atOffset: i];
    [score add: [lu getScore]];	/* This returns either an average or total,
				   depending on which method was called */
    [lu initialiseScore];	/* Set the score back to zero */
    [lu addToScore: [score call]];
				/* Add the score back as though it was a 
				   total */
    [score drop];
  }
  /* Now we make the scores cumulative */
  for(i = 1; i < [parameter nUses]; i++) {
    // Make the scores cumulative
    [[landUses atOffset: i]
      addToScore: [[[landUses atOffset: i - 1] getScore] call]];
  }
  /* Deal with the special case that all the scores are zero. Here, this will
     mean that all of the land uses have an equal chance of selection. */
  if([[[landUses atOffset: [parameter nUses] - 1] getScore]
       eq: [Interval constant: 0.0]]) {
    Interval *d = [Interval create];

    if([FearlusArguments debugCoreMethods]) {
      printf("B: %s: Buckets all contain zero. Constructing buckets so that "
	     "there is an equal chance of choosing each land use\n", 
	     sel_get_name(_cmd));
    }
    for([d set: [Interval constant: 1.0]], i = 0;
	   i < [parameter nUses];
	   i++, [d inc]) {
      [[landUses atOffset: i] setScore: [d call]];
    }
    [d drop];
  }
  /* Done dealing with the special case */
  // Make the choice
  choice = [uniformDblRand getDoubleWithMin: 0.0
			   withMax: [[[landUses
					atOffset:
					  [parameter nUses] - 1] getScore]
				      getMax]];
  // Find the land use that corresponds to the choice made, printing out a
  // table of the choices if required by the arguments to the program.
  if([FearlusArguments debugCoreMethods]) {
    printf("B: %s: Choice of land use made as follows\n", sel_get_name(_cmd));
    printf("\tLand Use      Score   Cumulative score range for choice\n");
  }
  lu = nil;
  for(i = 0; i < [parameter nUses]; i++) {
    if([FearlusArguments debugCoreMethods]) {
      Interval *lastScore = [Interval create];

      [lastScore set: ((i == 0) ? [Interval constant: 0.0]
		       : [[landUses atOffset: i - 1] getScore])];
      printf("\t%8u   %s   %s <= choice < %s",
	     [[landUses atOffset: i] getPIN],
	     [[[[landUses atOffset: i] getScore] diff: lastScore]
	       printfVar: "% 8g"],
	     [lastScore printfVar: "% 8g"],
	     [[[landUses atOffset: i] getScore] printfVar: "% 8g"]);
      [Interval printfclean];
      [lastScore drop];
    }
    if([[[landUses atOffset: i] getScore] ge: [Interval constant: choice]]
       && lu == nil) {
      lu = [landUses atOffset: i];
      if([FearlusArguments debugCoreMethods]) {
	printf(" <-- choice = %g", choice);
      }
      else {
	break;
      }
    }
    if([FearlusArguments debugCoreMethods]) {
      printf("\n");
    }
  }
  return lu;
}

/*

getDeterministicOptimumChoice

Return the land use with the highest score, with preference for land uses
with lower PINs.

*/

-(LandUse *)getDeterministicOptimumChoice {
  int i;
  LandUse *lu;

  lu = [landUses atOffset: 0];
  for(i = 1; i < [landUses getCount]; i++) {
    if([[[landUses atOffset: i] getScore] gt: [lu getScore]]) {
      lu = [landUses atOffset: i];
    }
    else if(([[[landUses atOffset: i] getScore] eq: [lu getScore]])
       && ([[landUses atOffset: i] getPIN] < [lu getPIN])) {
      lu = [landUses atOffset: i];
    }
  }
  if([FearlusArguments debugCoreMethods]) {
    printf("B: %s: Choice of land use made as follows\n", sel_get_name(_cmd));
    printf("\tLand Use      Score\n");
    for(i = 0; i < [landUses getCount]; i++) {
      printf("\t%8u   %s", [[landUses atOffset: i] getPIN],
	     [[[landUses atOffset: i] getScore] printfVar: "% 8g"]);
      [Interval printfclean];
      if([landUses atOffset: i] == lu) {
	printf(" <-- choice = (highest score then lowest ID)\n");
      }
      else {
	printf("\n");
      }
    }
  }
  return lu;
}

/*

getRandomOptimumChoice

Return the land use with the highest score, choosing at random between those
with the highest score when there is more than one.

*/

-(LandUse *)getRandomOptimumChoice {
  int i, choice;
  LandUse *lu;
  id options;

  options = [List create: scratchZone];
  lu = [landUses atOffset: 0];
  [options addLast: lu];
  for(i = 1; i < [landUses getCount]; i++) {
    if([[[landUses atOffset: i] getScore] gt: [lu getScore]]) {
      lu = [landUses atOffset: i];
      [options removeAll];
      [options addLast: lu];
    }
    else if([[[landUses atOffset: i] getScore] eq: [lu getScore]]) {
      [options addLast: [landUses atOffset: i]];
    }
  }
  choice = [uniformIntRand getIntegerWithMin: 0
			   withMax: [options getCount] - 1];
  lu = [options atOffset: choice];
  if([FearlusArguments debugCoreMethods]) {
    int j;

    printf("B: %s: Choice of land use made as follows\n", sel_get_name(_cmd));
    printf("\tLand Use      Score  Option\n");
    j = 0;
    for(i = 0; i < [landUses getCount]; i++) {
      printf("\t%8u   %s", [[landUses atOffset: i] getPIN],
	     [[[landUses atOffset: i] getScore] printfVar: "% 8g"]);
      [Interval printfclean];
      if([options contains: [landUses atOffset: i]]) {
	printf(" % 7d", ++j);		/* We're assuming the list contains
					   things in PIN order, here. This is
					   the order they were put in in the 
					   loop above, so that's OK. */
      }
      if([landUses atOffset: i] == lu) {
	printf(" <-- choice = %d\n", choice + 1);
      }
      else {
	printf("\n");
      }
    }
  }
  [options drop];
  return lu;
}

@end
