/*
    FEARLUS model0-7-1: SelectUseBucketRationals.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the SelectUseBucketRationals object. For now, the scores 
are held in the land use objects themselves.



*/

#import "SelectUseBucketRationals.h"
#import "FearlusArguments.h"
#import "LandUse.h"
#import "Parameter.h"
#import <DoubleRationalWarn.h>
#import "AssocArray.h"

@implementation SelectUseBucketRationals

/*

create:withParameters:andLandUses:

Return a freshly created SelectUseBucketRationals object. This will create 
a hash table with all the scores, using the land uses as keys.

*/

+create: aZone withParameters: (Parameter *)p andLandUses: lu {
  SelectUseBucketRationals *obj;
  int i;

  obj = [super create: aZone];
  obj->parameter = p;
  obj->landUses = lu;

  obj->table = [AssocArray createBegin: aZone];
  [obj->table setSize: (unsigned)[p nUses]];
  [obj->table createEnd];
  
  for(i = 0; i < (int)[p nUses]; i++) {
    DoubleRationalWarn *score = 
      [DoubleRationalWarn parseKeepSameClass: "DoubleRationalWarn=0!ALL"];
    [obj->table addObject: score withKey: [lu atOffset: i]];
  }

  return obj;
}

/*

addScore:toBucketForLandUse:

Add a score to the bucket for the specified land use.

*/

-(void)addScore: (DoubleRationalWarn *)s toBucketForLandUse: (LandUse *)lu {
  DoubleRationalWarn *currentScore =  [table getObjectWithKey: lu];
  [currentScore add: s];
  [s free];
}

/*

getScoreInBucketOfLandUse:

Return the score in the bucket of the specified land use.

*/

-(DoubleRationalWarn *)getScoreInBucketOfLandUse: (LandUse *)lu {
  return [[table getObjectWithKey: lu] copy];
}

/*

setScore:inBucketForLandUse:

Set a score to the bucket for the specified land use.

*/

-(void)setScore: (DoubleRationalWarn *)s inBucketForLandUse: (LandUse *)lu {
  DoubleRationalWarn *currentScore =  [table getObjectWithKey: lu];
  [currentScore set: s];
  [s free];
}


/*

getRandomOptimumChoice

Return the land use with the highest score, choosing at random between those
with the highest score when there is more than one.

*/

-(LandUse *)getRandomOptimumChoice {
  int i, choice;
  LandUse *lu;
  id options;

  options = [List create: scratchZone];
  lu = [landUses atOffset: 0];
  [options addLast: lu];
  for(i = 1; i < [landUses getCount]; i++) {
    if([[self getScoreInBucketOfLandUse: [landUses atOffset: i]] 
	 gt: [self getScoreInBucketOfLandUse: lu]]) {
      lu = [landUses atOffset: i];
      [options removeAll];
      [options addLast: lu];
    }
    else if([[self getScoreInBucketOfLandUse: [landUses atOffset: i]]
	      eq: [self getScoreInBucketOfLandUse: lu]]) {
      [options addLast: [landUses atOffset: i]];
    }
  }
  choice = [uniformIntRand getIntegerWithMin: 0
			   withMax: [options getCount] - 1];
  lu = [options atOffset: choice];
  if([FearlusArguments debugCoreMethods]) {
    int j;

    printf("B: %s: Choice of land use made as follows\n", sel_get_name(_cmd));
    printf("\tLand Use      Score  Option\n");
    j = 0;
    for(i = 0; i < [landUses getCount]; i++) {
      printf("\t%8u   ", [[landUses atOffset: i] getPIN]);
      [[self getScoreInBucketOfLandUse: [landUses atOffset: i]] write: stdout];
      if([options contains: [landUses atOffset: i]]) {
	printf(" % 7d", ++j);		/* We're assuming the list contains
					   things in PIN order, here. This is
					   the order they were put in in the 
					   loop above, so that's OK. */
      }
      if([landUses atOffset: i] == lu) {
	printf(" <-- choice = %d\n", choice + 1);
      }
      else {
	printf("\n");
      }
    }
  }
  [options drop];
  return lu;
}

@end
